using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Threading.Tasks;
using FarmmapsApi.Models;
using FarmmapsApi.Services;
using FarmmapsApiSamples.Models;
using Google.Apis.Upload;
using Microsoft.Extensions.Logging;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using static FarmmapsApi.Extensions;
using static FarmmapsApiSamples.Constants;

namespace FarmmapsApiSamples
{
    public class HerbicideService : ITestFlow
    {
        private readonly ILogger<HerbicideService> _logger;
        private readonly FarmmapsApiService _farmmapsApiService;

        private HerbicideAgent liberatorTarweAgent;

        public HerbicideService(ILogger<HerbicideService> logger, FarmmapsApiService farmmapsApiService)
        {
            _logger = logger;
            _farmmapsApiService = farmmapsApiService;

            liberatorTarweAgent = new HerbicideAgent()
            {
                Middel = "Liberator",
                Grondsoort = "Klei",
                ExtraProductType = "OS",
                ActieveStofGehalte = "diflu100+flufenacet400",
                MinDosis = 0.33f,
                MaxDosis = 0.6f,
                A = 0.007f,
                B = 0.25f,
                C = 1.05f,
                D = 97f,
                E = 3f,
                P = 1f,
                Gewas = "Tarwe",
                Opmerking = "Org. stof arme grond",
                TankMix = string.Empty
            };
        }

        public async Task TestFlow(List<UserRoot> roots)
        {
            var downloadFolder = "Downloads";
            if (!Directory.Exists(downloadFolder))
                Directory.CreateDirectory(downloadFolder);
            
            var myDrive = roots.SingleOrDefault(r => r.Name == "My drive");
            if (myDrive == null)
            {
                _logger.LogError("Could not find a needed root item");
                return;
            }

            var uploadedRoot = roots.SingleOrDefault(r => r.Name == "Uploaded");
            if (uploadedRoot == null)
            {
                _logger.LogError("Could not find a needed root item");
                return;
            }
            
            _logger.LogInformation("Creating cropfield");
            var cropfieldItem = await CreateCropfieldItemAsync(myDrive.Code);
            
            _logger.LogInformation("Uploading data");
            var dataItem = await UploadData(uploadedRoot, GEOTIFF_PROCESSED_ITEMTYPE, Path.Combine("Data", "Lutum.tiff"));
            if(dataItem == null)
            {
                _logger.LogError($"Failed to upload data");
                return;
            }

            _logger.LogInformation("Calculating appliance map");
            var applianceMapItem = await CalculateApplianceMap(cropfieldItem, dataItem, liberatorTarweAgent);
            if (applianceMapItem == null)
            {
                return;
            }
            
            _logger.LogInformation("Downloading appliance map");
            await _farmmapsApiService.DownloadItemAsync(applianceMapItem.Code, Path.Combine(downloadFolder, $"{applianceMapItem.Name}.zip"));
        }

        private async Task<Item> CreateCropfieldItemAsync(string parentItemCode)
        {
            var currentYear = new DateTime(DateTime.UtcNow.Year, 1, 1);
            var cropfieldItemRequest = new ItemRequest()
            {
                ParentCode = parentItemCode,
                ItemType = CROPFIELD_ITEMTYPE,
                Name = "Cropfield VRA Herbicide LUTUM",
                DataDate = currentYear,
                DataEndDate = currentYear.AddMonths(3),
                Data = JObject.Parse("{}"),
                Geometry = JObject.Parse(
                    @"{""type"":""Polygon"",""coordinates"":[[[4.617786844284247,52.22533706956424],[4.618642601314543,52.225938364585989],[4.6192153806397,52.22563988897754],[4.619192414656403,52.2256242822442],[4.620306732153958,52.225031745661528],[4.620542019225217,52.22519855319158],[4.621157509147853,52.22487436515405],[4.623387917230182,52.22367660757213],[4.624563444939009,52.22304740241544],[4.624562779355982,52.223046635247019],[4.624534908813479,52.22302596787506],[4.627873021330343,52.221240670658399],[4.627504935938338,52.220104419135129],[4.627324878706837,52.22020569669098],[4.627320696113512,52.22020660117888],[4.626707169518044,52.22053923770041],[4.624700376420229,52.221619047547488],[4.623471571183885,52.22227447969577],[4.623471511010673,52.22227500174403],[4.623468838689317,52.22228052566992],[4.617786844284247,52.22533706956424]]]}")
            };

            return await _farmmapsApiService.CreateItemAsync(cropfieldItemRequest);
        }

        private async Task<Item> UploadData(UserRoot root, string itemType, string filePath)
        {
            var startUpload = DateTime.UtcNow;
            var result = await _farmmapsApiService.UploadFile(filePath, root.Code,
                progress => _logger.LogInformation($"Status: {progress.Status} - BytesSent: {progress.BytesSent}"));

            if (result.Progress.Status == UploadStatus.Failed)
                return null;

            var fileName = Path.GetFileNameWithoutExtension(filePath);
            return await FindItem(root.Code, itemType, fileName, 
                i => i.Created >= startUpload && 
                     i.Name.ToLower().Contains(fileName.ToLower()));
        }

        private async Task<Item> CalculateApplianceMap(Item cropfieldItem, Item inputItem, HerbicideAgent agent)
        {
            var taskRequest = new TaskRequest()
            {
                TaskType = "vnd.farmmaps.task.vraherbicide"
            };
            taskRequest.attributes["inputCode"] = inputItem.Code;
            taskRequest.attributes["agent"] = JsonConvert.SerializeObject(agent);

            var taskCode = await _farmmapsApiService.QueueTaskAsync(cropfieldItem.Code, taskRequest);
            await PollTask(TimeSpan.FromSeconds(3), async (tokenSource) =>
            {
                _logger.LogInformation("Checking vraherbicide task status");
                var itemTaskStatus = await _farmmapsApiService.GetTaskStatusAsync(cropfieldItem.Code, taskCode);
                if (itemTaskStatus.State == ItemTaskState.Error || itemTaskStatus.State == ItemTaskState.Ok)
                    tokenSource.Cancel();
            });
            
            var itemTask = await _farmmapsApiService.GetTaskStatusAsync(cropfieldItem.Code, taskCode);
            if (itemTask.State == ItemTaskState.Error)
            {
                _logger.LogError($"Something went wrong with task execution: {itemTask.Message}");
                return null;
            }

            var itemName = $"VRAHerbicide {agent.Middel}";
            var applianceMapItem = await FindItem(cropfieldItem.Code,
                GEOTIFF_PROCESSED_ITEMTYPE, itemName, 
                i => i.Updated >= itemTask.Finished.GetValueOrDefault(DateTime.UtcNow) &&
                            i.Name.ToLower().Contains(itemName.ToLower()));
            
            if (applianceMapItem == null)
            {
                _logger.LogError("Could not find the VRAHerbicide geotiff child item under cropfield");
                return null;
            }

            return applianceMapItem;
        }

        private async Task<Item> FindItem(string parentCode, string itemType, string containsName,
            Func<Item,bool> filter = null, int maxTries = 10)
        {
            Item dataItem = null;
            int tries = 0;
            await PollTask(TimeSpan.FromSeconds(3), async source =>
            {
                _logger.LogInformation($"Trying to get {containsName} data");
                var uploadedFilesChildren = await _farmmapsApiService.GetItemChildrenAsync(parentCode, itemType);
                if (uploadedFilesChildren.Count > 0)
                {
                    Func<Item, bool> func = filter ?? (i => i.Name.ToLower().Contains(containsName.ToLower()));
                    dataItem = uploadedFilesChildren.FirstOrDefault(func);
                    source.Cancel();
                } else if (tries == maxTries)
                {
                    source.Cancel();
                }

                tries++;
            });

            if (dataItem == null)
            {
                _logger.LogError("dataItem not found");
                return null;
            }
            
            _logger.LogInformation($"Found {containsName} item");
            return dataItem;
        }
    }
}